/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.descriptor.extension;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;

import org.ow2.petals.jbi.descriptor.AbstractJAXBSingleton;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.extension.exception.JbiExtensionException;
import org.ow2.petals.jbi.descriptor.extension.exception.NoComponentNameDeployableServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.exception.NoIdentificationDeployableServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.exception.NotDeployableServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.exception.NotServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.exception.SeveralComponentNameDeployableServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.exception.SeveralIdentificationDeployableServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.generated.Identification;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.Services;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;

/**
 * This Class defines the a JBI Descriptor builder for the Petals extensions of the JBI descriptor.
 * <p>
 * This class is thread-safe.
 * </p>
 * 
 * @author Christophe DENEUX - Linagora
 */
public class JBIDescriptorExtensionBuilder extends AbstractJAXBSingleton {

    /**
     * JBI schema name of Petals extensions
     */
    public static final String XSD_PETALS_JBI = "petals-jbi.xsd";

    /**
     * Local name of the tag 'identification' of a deployable service unit
     */
    public static final String TAG_DEPLOYABLE_SU_IDENTIFICATION = "identification";

    /**
     * Local name of the tag 'target component' of a deployable service unit
     */
    public static final String TAG_DEPLOYABLE_SU_TARGET_COMPONENT = "component-name";

    /**
     * Petals JBI extension namespace
     */
    public static final String PETALS_JBI_NAMESPACE = "http://petals.ow2.org/extensions/jbi/1.0";

    private static JBIDescriptorException INSTANCE_EXCEPTION;

    private static JBIDescriptorExtensionBuilder INSTANCE;

    public static synchronized JBIDescriptorExtensionBuilder getInstance() throws JBIDescriptorException {
        if (INSTANCE_EXCEPTION != null) {
            throw INSTANCE_EXCEPTION;
        }
        if (INSTANCE == null) {
            try {
                INSTANCE = new JBIDescriptorExtensionBuilder();
            } catch (JBIDescriptorException e) {
                INSTANCE_EXCEPTION = e;
                throw INSTANCE_EXCEPTION;
            }
        }
        return INSTANCE;
    }

    protected JBIDescriptorExtensionBuilder() throws JBIDescriptorException {
        try {
            super.init(Identification.class, JBIDescriptorBuilder.XSD_JBI, XSD_PETALS_JBI);
        } catch (final IOException e) {
            throw new JBIDescriptorException("Failed to get resources from the current class-path", e);
        } catch (final SAXException e) {
            throw new JBIDescriptorException("Failed to parse resources", e);
        } catch (JAXBException e) {
            throw new JBIDescriptorException("Failed to create JAXB objects", e);
        }
    }

    /**
     * Extract the identification part of a deployable service unit
     * 
     * @param jbi
     *            The JBI descriptor in which the deployable service unit identification part must be extracted. Not
     *            <code>null</code>.
     * @return The {@link Identification} of a deployable service unit
     * @throws NotServiceUnitException
     *             The JBI descriptor is not the JBI descriptor of a service unit
     * @throws NotDeployableServiceUnitException
     *             The JBI descriptor is not the JBI descriptor of a deployable service unit
     * @throws SeveralIdentificationDeployableServiceUnitException
     *             Several identification sections exist in the JBI descriptor of a deployable service unit
     * @throws NoIdentificationDeployableServiceUnitException
     *             No identification section exists in the JBI descriptor of a deployable service unit
     * @throws JbiExtensionException
     */
    public Identification getDeployableServiceUnitIndentification(final Jbi jbi) throws JbiExtensionException {

        final Services services = jbi.getServices();
        if (services == null) {
            throw new NotServiceUnitException();
        }

        final List<Element> elements = services.getAnyOrAny();
        if (elements == null || elements.isEmpty()) {
            throw new NotDeployableServiceUnitException();
        }

        Identification identification = null;
        for (final Element element : elements) {
            if (PETALS_JBI_NAMESPACE.equals(element.getNamespaceURI())
                    && TAG_DEPLOYABLE_SU_IDENTIFICATION.equals(element.getLocalName())) {
                if (identification == null) {
                    try {
                        final JAXBElement<Identification> root;
                        // The default Xerces unmarshaller is not thread safe
                        synchronized (getUnmarshaller()) {
                            root = getUnmarshaller().unmarshal(element, Identification.class);
                        }

                        identification = root.getValue();

                    } catch (final JAXBException e) {
                        throw new JbiExtensionException("Failed to build the deployable service unit identification", e);
                    }
                } else {
                    throw new SeveralIdentificationDeployableServiceUnitException();
                }
            }
        }
        
        if (identification == null) {
            throw new NoIdentificationDeployableServiceUnitException();
        }

        return identification;
    }

    /**
     * Set or override the identification part of a deployable service unit
     * 
     * @param jbi
     *            The JBI descriptor in which the identification part must be added
     * @param suIdentification
     *            The deployable service unit identification part to add
     * @throws JbiExtensionException
     */
    public void setDeployableServiceUnitIdentification(final Jbi jbi, final Identification suIdentification)
            throws JbiExtensionException {

        // Check that the JBI descriptor against the number of identification parts
        Identification existingIdent;
        try {
            existingIdent = getDeployableServiceUnitIndentification(jbi);
        } catch (final NotDeployableServiceUnitException e) {
            // No extension about deployable service unit already exists in the JBI descriptor.
            // We will add it.
            existingIdent = null;
        } catch (final NoIdentificationDeployableServiceUnitException e) {
            // No identification part already exists in the JBI decriptor, we will add it
            existingIdent = null;
        }

        if (existingIdent == null) {
            // We add the identification part
            final Document suIdentDoc = DocumentBuilders.newDocument();
            marshallIdentification(suIdentification, suIdentDoc);
            jbi.getServices().getAnyOrAny().add(suIdentDoc.getDocumentElement());
        } else {
            // We update the identification part
            final Document suIdentDoc = DocumentBuilders.newDocument();
            marshallIdentification(suIdentification, suIdentDoc);
            final Iterator<Element> itElements = jbi.getServices().getAnyOrAny().iterator();
            while (itElements.hasNext()) {
                final Element element = itElements.next();
                if (PETALS_JBI_NAMESPACE.equals(element.getNamespaceURI())
                        && TAG_DEPLOYABLE_SU_IDENTIFICATION.equals(element.getLocalName())) {

                    itElements.remove();
                    jbi.getServices().getAnyOrAny().add(suIdentDoc.getDocumentElement());
                    break;
                }
            }
        }
    }

    /**
     * Write the deployable service unit identification part into an DOM {@link Document}
     * 
     * @param suIdentification
     *            The deployable service unit identification part to add
     * @param docIdentification
     *            The target DOM {@link Document} in which the deployable service unit identification part will be
     *            marshaled
     */
    private void marshallIdentification(final Identification suIdentification,
            final Document docIdentification) throws JbiExtensionException {

        try {
            synchronized (getMarshaller()) {
                getMarshaller().marshal(suIdentification, docIdentification);
            }
        } catch (JAXBException ex) {
            throw new JbiExtensionException("Failed to write the deployable service unit identification part", ex);
        }
    }

    /**
     * Extract the target component name part of a deployable service unit
     * 
     * @param jbi
     *            The JBI descriptor in which the deployable service unit target component part must be extracted. Not
     *            <code>null</code>.
     * @return The target component name of a deployable service unit
     * @throws NotServiceUnitException
     *             The JBI descriptor is not the JBI descriptor of a service unit
     * @throws NotDeployableServiceUnitException
     *             The JBI descriptor is not the JBI descriptor of a deployable service unit
     * @throws SeveralComponentNameDeployableServiceUnitException
     *             Several target component name sections exist in the JBI descriptor of a deployable service unit
     * @throws NoComponentNameDeployableServiceUnitException
     *             No target component name section exists in the JBI descriptor of a deployable service unit
     */
    public String getDeployableServiceUnitTargetComponent(final Jbi jbi) throws JbiExtensionException {

        final Services services = jbi.getServices();
        if (services == null) {
            throw new NotServiceUnitException();
        }

        final List<Element> elements = services.getAnyOrAny();
        if (elements == null || elements.isEmpty()) {
            throw new NotDeployableServiceUnitException();
        }

        String targetComponentName = null;
        for (final Element element : elements) {
            if (PETALS_JBI_NAMESPACE.equals(element.getNamespaceURI())
                    && TAG_DEPLOYABLE_SU_TARGET_COMPONENT.equals(element.getLocalName())) {
                if (targetComponentName == null) {
                    targetComponentName = element.getTextContent();
                } else {
                    throw new SeveralComponentNameDeployableServiceUnitException();
                }
            }
        }
        
        if (targetComponentName == null || targetComponentName.trim().isEmpty()) {
            throw new NoComponentNameDeployableServiceUnitException();
        }

        return targetComponentName;
    }

    /**
     * Set or override the target component part of a deployable service unit
     * 
     * @param jbi
     *            The JBI descriptor in which the identification part must be added
     * @param targetComponentName
     *            The deployable service unit target component part to add
     * @throws JbiExtensionException
     */
    public void setDeployableServiceUnitTargetComponent(final Jbi jbi, final String targetComponentName)
            throws JbiExtensionException {

        // Check that the JBI descriptor against the number of identification parts
        String existingTargetComponent;
        try {
            existingTargetComponent = getDeployableServiceUnitTargetComponent(jbi);
        } catch (final NotDeployableServiceUnitException e) {
            // No extension about deployable service unit already exists in the JBI descriptor.
            // We will add it.
            existingTargetComponent = null;
        } catch (final NoComponentNameDeployableServiceUnitException e) {
            // No target component part already exists in the JBI decriptor, we will add it
            existingTargetComponent = null;
        }

        if (existingTargetComponent == null) {
            // We add the target component part
            final Document suDoc = DocumentBuilders.newDocument();
            final Element targetComponentElt = suDoc.createElementNS(PETALS_JBI_NAMESPACE,
                    TAG_DEPLOYABLE_SU_TARGET_COMPONENT);
            targetComponentElt.setTextContent(targetComponentName);
            suDoc.appendChild(targetComponentElt);
            jbi.getServices().getAnyOrAny().add(suDoc.getDocumentElement());
        } else {
            // We update the target component part
            for (final Element element : jbi.getServices().getAnyOrAny()) {
                if (PETALS_JBI_NAMESPACE.equals(element.getNamespaceURI())
                        && TAG_DEPLOYABLE_SU_TARGET_COMPONENT.equals(element.getLocalName())) {

                    element.setTextContent(targetComponentName);
                    break;
                }
            }
        }
    }
}
